"""
A module for lazy loading Prefect Blocks.

The `@lazy_load` decorator is centered around a model to attach blocks to. Each block
is defined with two parts: a block name and a block property. The block name is a
string that identifies the block. The block property is a property that loads the block
on first access. The block property should be decorated with the `@lazy_load` decorator,
and the block's type should be specified in the property annotations.

Example:

```python
from prefect.blocks.system import Secret
from prefecto.blocks import lazy_load


class Blocks:
    \"""Class for lazy loading Prefect Blocks.\"""

    # Define the block name variables
    password_block: str = "secret-password"

    @property
    @lazy_load("password_block")
    def password(self) -> Secret:
        \"""The password block.\"""

blocks = Blocks()
password = blocks.password
print(password)
# Secret(value=SecretStr('**********'))
print(password.get())
# my-secret-password$123
```

This technique is useful for preventing the blocks from loading until they are actually
needed. This can be important during unit testing, where connections to the Prefect
server may not exist. The alternative would be to call `block.load()` directly, which
would load the block every time it is called and could lead to performance issues.

The loader can be integrated with `pydantic-settings` to load block names from
environment variables. This can be achieved by setting the block name variables
as class variables with default values set to the environment variables.

Example:

```python
from pydantic_settings import BaseSettings, SettingsConfigDict
from prefect.blocks.system import Secret
from prefecto.blocks import lazy_load


class Blocks(BaseSettings):
    \"""Class for lazy loading Prefect Blocks.\"""

    model_config = SettingsConfigDict(env_prefix="BLOCKS_")

    # Define the block name variables
    password_block: str = "secret-password"

    @property
    @lazy_load("password_block")
    def password(self) -> Secret:
        \"""The password block.\"""

```

"""

import asyncio
import functools
import types
from typing import Union, get_args, get_origin

from prefect._internal.concurrency.api import from_sync
from prefect.blocks.abstract import Block
from prefect.utilities.asyncutils import sync_compatible


@sync_compatible
async def load_block(block_type: type[Block], block_name: str) -> Block:
    """Load a block.

    Args:
        block_type (type[Block]): The block type.
        block_name (str): The block name.
    """
    # Sometimes when calling this load func in an async context,
    # block_type.load() will return a coroutine object instead of the block.
    # This is a workaround to handle that case.
    block = block_type.load(block_name)
    return (await block) if asyncio.iscoroutine(block) else block


def lazy_load(varname: str):
    """Decorator for lazy loading a block.

    Args:
        varname (str): The variable name of the block name var.

    Returns:
        The decorator for the loader.

    Example:

    Specify the block name, then create a decorated property to load the block on
    first access.

    ```python
    class MyClass:
        block_name = "block_name"

        @property
        @lazy_load("block_name")
        def block(self) -> BlockType:
            \"""The block.\"""
    ```
    """

    def decorator(func):
        """Decorator for the loader."""

        block_type = func.__annotations__["return"]

        if (get_origin(block_type) is Union) or isinstance(block_type, types.UnionType):
            raise TypeError(
                f"Only one block type can be specified for a block property. Received: {get_args(block_type)}",
            )
        elif isinstance(block_type, type) and not issubclass(block_type, Block):
            raise TypeError(
                f"The block type must be a subclass of Block. Received: {block_type}",
            )

        block_varname = f"_{func.__name__}_block"

        @functools.wraps(func)
        def lazy_property(self):
            """Lazy property loader for a block."""
            block = getattr(self, block_varname, None)
            if not block:
                block = load_block(block_type, getattr(self, varname))
                if asyncio.iscoroutine(block):
                    # This is a workaround for a case where the block is first accessed
                    # within an asynchronous flow. In that case, the @sync_compatible
                    # decorator will return a coroutine object instead of the block.
                    # This will run the coroutine in the loop thread to get the block.
                    block = from_sync.call_in_loop_thread(lambda: block)
                setattr(self, block_varname, block)
            return block

        lazy_property.varname = varname
        lazy_property.block_type = block_type
        lazy_property.block_varname = block_varname

        return lazy_property

    return decorator



"""
Tools to improve Prefect states.

"""

from __future__ import annotations

from prefect import states


def is_terminal(state: states.State) -> bool:
    """Return True if the state is terminal. Terminal states are:

    - Cancelled
    - Completed
    - Crashed
    - Failed
    """
    TERMINALS = [
        state.is_cancelled,
        state.is_completed,
        state.is_crashed,
        state.is_failed,
    ]
    for terminal in TERMINALS:
        if terminal():
            return True
    return False



import re
from typing import Union
from unittest.mock import patch

import pytest
from prefect.blocks.system import Secret

from prefecto.blocks import lazy_load, load_block


def test_load_block():

    with patch(
        "prefect.blocks.system.Secret.load", return_value=Secret(value="abc-123")
    ):
        block = load_block(Secret, "block")
        assert isinstance(block, Secret)
        assert block.get() == "abc-123"


def test_lazy_load():

    with patch(
        "prefect.blocks.system.Secret.load", return_value=Secret(value="abc-123")
    ):

        class Blocks:
            password = "block"

            @property
            @lazy_load("password")
            def pw(self) -> Secret:
                return load_block(Secret, getattr(self, "password"))

        blocks = Blocks()
        assert isinstance(blocks.pw, Secret)
        assert blocks.pw.get() == "abc-123"


def test_lazy_load_fail_on_union_return():

    with pytest.raises(
        TypeError,
        match=re.escape(
            "Only one block type can be specified for a block property. Received: (<class 'prefect.blocks.system.Secret'>, <class 'str'>)",
        ),
    ):

        class _:
            password = "block"

            @property
            @lazy_load("password")
            def pw(self) -> Secret | str:
                return load_block(Secret, getattr(self, "password"))

    with pytest.raises(
        TypeError,
        match=re.escape(
            "Only one block type can be specified for a block property. Received: (<class 'prefect.blocks.system.Secret'>, <class 'str'>)",
        ),
    ):

        class _:
            password = "block"

            @property
            @lazy_load("password")
            def pw(self) -> Union[Secret, str]:
                return load_block(Secret, getattr(self, "password"))


@pytest.mark.asyncio
async def test_lazy_load_run_coro_return_val():
    with patch(
        "prefect.blocks.system.Secret.load", return_value=Secret(value="abc-123")
    ):

        class Blocks:
            password = "block"

            @property
            @lazy_load("password")
            def pw(self) -> Secret:
                return load_block(Secret, getattr(self, "password"))

        blocks = Blocks()
        assert isinstance(blocks.pw, Secret)
        assert blocks.pw.get() == "abc-123"



"""
Kill switch logic classes for stopping the execution of a
[`BatchTask`][src.prefecto.concurrency.batch_task.BatchTask].

"""

import abc

from prefect.states import State


class KillSwitchError(Exception):
    """Error raised when a kill switch is activated."""

    def __init__(self, message: str, ks: "KillSwitch"):
        super().__init__(message)
        self.ks = ks


class KillSwitch(abc.ABC):
    """Abstract base class for a kill switch.

    Each invocation of the `should_flip_switch` method should advance the state of the
    kill switch and return if the kill switch should be activated. The
    `raise_if_triggered` method should raise a `KillSwitchError` if the kill switch has
    been activated.

    ```python

    class AnyFailedSwitch(KillSwitch):
        def should_flip_switch(self, state: State) -> bool:
            return state.is_failed() or state.is_crashed()

        def raise_if_triggered(self, state: State):
            if self.should_flip_switch(state):
                raise KillSwitchError("Failed task detected.", self)
    ```
    """

    @abc.abstractmethod
    def should_flip_switch(self, state: State) -> bool:
        """Check if this state should flip the kill switch.

        Returns:
            `True` if the kill switch should be activated, `False` otherwise.

        """

    @abc.abstractmethod
    def raise_if_triggered(self, state: State):
        """Check a state and raise a `KillSwitchError` if the kill switch has been activated.

        Raises:
            KillSwitchError: If the kill switch has been activated.

        """


class AnyFailedSwitch(KillSwitch):
    """A kill switch that activates if any task fails."""

    def should_flip_switch(self, state: State) -> bool:
        """Check if the state is failed or crashed."""
        return state.is_failed() or state.is_crashed()

    def raise_if_triggered(self, state: State):
        """Raise a `KillSwitchError` if the state is failed or crashed."""
        if self.should_flip_switch(state):
            raise KillSwitchError("Failed task detected.", self)


class CountSwitch(KillSwitch):
    """A kill switch that activates after a certain number of tasks fail.

    Args:
        count (int): The number of failed or crashed states that should trigger the kill
            switch.

    """

    def __init__(self, max_count: int):
        self.max_count = max_count
        self._current_count = 0

    def should_flip_switch(self, state: State) -> bool:
        """Increment the count if the state is failed or crashed and return if the count exceeds
        the maximum.
        """
        if state.is_failed() or state.is_crashed():
            self._current_count += 1
        return self._current_count >= self.max_count

    def raise_if_triggered(self, state: State):
        """Raise a `KillSwitchError` if the count exceeds the maximum."""
        if self.should_flip_switch(state):
            raise KillSwitchError(f"{self.max_count} failed tasks detected.", self)


class RateSwitch(KillSwitch):
    """A kill switch that activates after the failure rate exceeds a certain threshold.
    Requires a minimum number of states to sample.

    Args:
        min_sample (int): The minimum number of states to sample.
        max_fail_rate (float): The maximum frequency of failed or crashed states.

    """

    def __init__(self, min_sample: int, max_fail_rate: float):
        self.min_sample = min_sample
        self.max_fail_rate = max_fail_rate
        self._current_count = 0
        self._failed_count = 0

    def should_flip_switch(self, state: State) -> bool:
        """Increment the count if the state is failed or crashed and return if the failure rate
        equals or exceeds the max rate.
        """
        self._current_count += 1
        if state.is_failed() or state.is_crashed():
            self._failed_count += 1
        return (
            self._current_count >= self.min_sample
            and self._failed_count / self._current_count >= self.max_fail_rate
        )

    def raise_if_triggered(self, state: State):
        """Raise a `KillSwitchError` if the failure rate equals or exceeds the maximum rate."""
        if self.should_flip_switch(state):
            raise KillSwitchError(
                f"Failure rate exceeded {self.max_fail_rate} after {self.min_sample} samples.",
                self,
            )



"""
Module to improve Prefect concurrency operations.
"""

from __future__ import annotations

from typing import TypeVar

from prefect import unmapped
from prefect.futures import PrefectFuture
from prefect.tasks import Task
from prefect.utilities.callables import get_call_parameters
from typing_extensions import ParamSpec

from prefecto import logging, states
from prefecto.concurrency.kill_switch import KillSwitch

T = TypeVar("T")  # Generic type var for capturing the inner return type of async funcs
R = TypeVar("R")  # The return type of the user's function
P = ParamSpec("P")  # The parameters of the task
P_i = ParamSpec("P_i")  # The type of each parameter

MapArgument = list[P_i] | unmapped[P_i]
Batch = dict[str, MapArgument]


class BatchTask:
    """Wraps a `Task` to perform `Task.map` in batches, reducing the number of
    concurrent tasks, mellowing Prefect API requests, and allowing for faster
    failure detection.

    ```python
    from prefect import flow, task
    from prefecto.concurrency import BatchTask

    @task
    def add(a, b):
        return a + b

    @flow
    def my_flow():
        batch_add = BatchTask(add, 2)
        return batch_add.map([1,2,3,4], [2,3,4,5])

    ```

    The `kill_switch` argument can be used to stop the execution of the task
    after a certain condition is met.

    ```python
    from prefect import flow, task
    from prefecto.concurrency import BatchTask, AnyFailedSwitch

    @task
    def add(a, b):
        return a + b

    @flow
    def my_flow():
        batch_add = BatchTask(add, 2, kill_switch=AnyFailedSwitch())
        # This will error on the first batch and stop the execution.
        return batch_add.map([1,2,3,4], [1,'2',3,4])

    ```

    See [kill switches][src.prefecto.concurrency.kill_switch] for more information.

    """

    def __init__(
        self, task: Task[P, R], size: int, kill_switch: KillSwitch | None = None
    ):
        """Wrap the `task` to be executed in batches of `size`.

        Args:

                task (Task): The task to wrap.
                size (int): The size of the batches to perform `Task.map` on.
                kill_switch (KillSwitch, optional): A kill switch to stop the execution of the task
                    after a certain condition is met.
        """
        self.task: Task = task
        self.size: int = size

        if kill_switch is not None and not isinstance(kill_switch, KillSwitch):
            raise TypeError(
                f"Expected 'kill_switch' to be a subclass of 'KillSwitch', got {type(kill_switch)}."
            )
        self._kill_switch = kill_switch

    def _make_batches(self, **params: MapArgument) -> list[Batch]:
        """Create batches of arguments to pass to the `Task.map` calls.

        Args:
            **params (MapArgument): Keyword arguments where each value is an
            iterable of equal length or an `unmapped` object. Should be at least
            one non-`unmapped` argument.

        Returns:
            (list[dict[str, list[P_i] | unmapped[P_i]]]): A list of dictionaries where each
            dictionary has the same keys as the provided keyword arguments.
            The values of the dictionaries are lists with lengths no greater
            than `BatchTask.size`.

        Examples:

            ```python
            BatchTask(task, 3)._make_batches(a=[1,2,3,4,5], b=[2,3,4,5,6])
            ```
            ```json
            [
                {"a": [1,2,3], "b": [2,3,4]},
                {"a": [4,5], "b": [4,5,6]}
            ]
            ```
        """
        parameters = sorted(params.keys())
        if len(parameters) == 0:
            raise ValueError("Must provide at least one iterable.")

        # Validate all are iterables
        for k in parameters:
            if not hasattr(params[k], "__iter__"):
                raise ValueError(f"Expected '{k}' to be an iterable.")

        # Get the first non-unmapped iterable
        length: int
        for j, name in enumerate(parameters):
            if not isinstance(params[name], unmapped):
                length = len(params[name])
                break
        else:
            # Logically, at least one iterable must be provided
            raise ValueError("Must provide at least one non-unmapped iterable.")

        # Assure all of equal length
        if len(parameters[j:]) > 1:
            for k in parameters[j + 1 :]:
                if not isinstance(params[k], unmapped) and not len(params[k]) == length:
                    raise ValueError(
                        f"Expected all iterables to be of length {length} like "
                        f"'{parameters[j]}'. '{k}' is length {len(params[k])}."
                    )

        batches = []

        i = 0
        for i in range(length // self.size):
            batch = {p: [] for p in parameters}
            for p in parameters:
                if isinstance(params[p], unmapped):
                    # Pass the unmapped argument to be handled by the task
                    batch[p] = params[p]
                else:
                    batch[p] = params[p][i * self.size : (i + 1) * self.size]
            batches.append(batch)

        # Add the remainder if there is one
        if length % self.size != 0:
            batch = {p: [] for p in parameters}
            for p in parameters:
                if isinstance(params[p], unmapped):
                    # Pass the unmapped argument to be handled by the task
                    batch[p] = params[p]
                else:
                    batch[p] = params[p][(length // self.size) * self.size :]
            batches.append(batch)

        return batches

    def map(self, *args: MapArgument, **kwds: MapArgument) -> list[PrefectFuture]:
        """Perform a `Task.map` operation in batches of the keyword arguments. The
        arguments must be iterables of equal length.

        Args:

            *args: Positional arguments to pass to the task.
            **kwds: Keyword arguments to pass to the task.

        Returns:
            A list of futures for each batch.

        Examples:

            ```python
            from prefect import flow, task
            from prefecto.concurrency import BatchTask

            @task
            def add(a, b):
                return a + b

            @flow
            def my_flow():
                batch_add = BatchTask(add, 2)
                return batch_add.map([1,2,3,4], [2,3,4,5])

            print(my_flow())
            ```

            ```log
            $ python my_flow.py
            01:31:51.012 | INFO    | prefect.engine - Created flow run 'beryl-moth' for flow 'test'
            01:31:52.238 | DEBUG   | Flow run 'beryl-moth' - Mapping 'add' batch 1 of 2.
            01:31:52.239 | INFO    | Flow run 'beryl-moth' - Created task run 'add-0' for task 'add'
            01:31:52.240 | INFO    | Flow run 'beryl-moth' - Submitted task run 'add-0' for execution.
            01:31:52.253 | INFO    | Flow run 'beryl-moth' - Created task run 'add-1' for task 'add'
            01:31:52.254 | INFO    | Flow run 'beryl-moth' - Submitted task run 'add-1' for execution.
            01:31:52.259 | DEBUG   | Flow run 'beryl-moth' - Mapping 'add' batch 2 of 2.
            01:31:52.258 | INFO    | Flow run 'beryl-moth' - Created task run 'add-3' for task 'add'
            01:31:52.258 | INFO    | Flow run 'beryl-moth' - Submitted task run 'add-3' for execution.
            01:31:52.260 | INFO    | Flow run 'beryl-moth' - Created task run 'add-2' for task 'add'
            01:31:52.261 | INFO    | Flow run 'beryl-moth' - Submitted task run 'add-2' for execution.
            01:31:52.675 | INFO    | Task run 'add-1' - Finished in state Completed()
            01:31:52.770 | INFO    | Task run 'add-0' - Finished in state Completed()
            01:31:52.885 | INFO    | Task run 'add-2' - Finished in state Completed()
            01:31:53.075 | INFO    | Task run 'add-3' - Finished in state Completed()
            01:31:53.979 | INFO    | Flow run 'beryl-moth' - Finished in state Completed()
            ```

            ```json
            [3, 5, 7, 9]
            ```
        """
        parameters = get_call_parameters(self.task.fn, args, kwds, apply_defaults=False)
        batches = self._make_batches(**parameters)

        return self._map(batches)

    def _map(self, batches: list[Batch]) -> list[PrefectFuture]:
        """Applies `Task.map` to each batch.

        Args:
            batches (list[Batch]): Batches of arguments to pass to
            `Task.map`.

        Returns:
            A list of futures for each batch.
        """
        if len(batches) == 0:
            return []
        logger = logging.get_prefect_or_default_logger()
        results: list[PrefectFuture] = []
        for i, batch in enumerate(batches[:-1]):
            logger.debug(f"Mapping {self.task.name} batch {i+1} of {len(batches)}.")
            # Map the batch
            futures = self.task.map(**batch)
            results.extend(futures)
            # Poll futures to ensure they are not active.
            is_processing: bool = True
            while is_processing:
                # If any future is still processing, loop and poll again.
                # If all futures are terminal, break while loop and continue
                if all(states.is_terminal(f.get_state()) for f in futures):
                    is_processing = False

            if self._kill_switch is not None:
                for f in futures:
                    self._kill_switch.raise_if_triggered(f.get_state())

        # Map the last batch
        logger.debug(
            f"Mapping {self.task.name} batch {len(batches)} of {len(batches)}."
        )
        results.extend(self.task.map(**batches[-1]))
        return results



"""
Opinionated Snowflake cursor for logging Snowflake execution calls with parameter obfuscation options.

```python
from snowflake.connector import connect
from prefecto.ext.snowflake import LogCursor

with connect(...).cursor(LogCursor) as cursor:
    cursor.execute(
        "SELECT * FROM table WHERE id = %(id)s AND secret = %(secret)s",
        params={"id": 123, "secret": "shhh"},
        obfuscate_params=["secret"],
        command_id="secret-selection",
    )
```

```
INFO - [secret-selection] Beginning command.
DEBUG - [secret-selection] Executing command:
SELECT * FROM table WHERE id = 123 AND secret = ****
INFO - [secret-selection] Command executed successfully.
```

"""

from __future__ import annotations

import logging
import sys
from typing import IO, Any, Sequence

import coolname
from snowflake.connector.cursor import SnowflakeCursor as _SnowflakeCursor
from snowflake.connector.file_transfer_agent import SnowflakeProgressPercentage

__python_version__ = sys.version_info


class CommandLogAdapter(logging.LoggerAdapter):
    """A logging adapter that prepends the command ID to the log messages."""

    def process(self, msg: str, kwargs: dict[str, Any]) -> tuple[str, dict[str, Any]]:
        return f"[{self.extra['command_id']}] {msg}", kwargs


def _obfuscate_params(
    params: dict[str, Any], obfuscate_params: bool | str | list[str]
) -> dict[str, Any]:
    """Obfuscates the parameters in the dictionary.

    Args:
        params (dict[str, Any]): The parameters to obfuscate.
        obfuscate_params (bool | str | list[str]): Whether to obfuscate the parameters.
            If `True`, obfuscates all parameters. If a string or list of strings, obfuscates only the specified parameters.

    Returns:
        dict[str, Any]: The obfuscated parameters.

    Example:

        ```python
        _obfuscate_params({"id": 123, "secret": "shhh"}, True)
        # {"id": "****", "secret": "****"}
        ```

    """
    if obfuscate_params is True:
        return {k: "****" for k in params}
    elif isinstance(obfuscate_params, str):
        return {k: "****" if k == obfuscate_params else v for k, v in params.items()}
    elif isinstance(obfuscate_params, list):
        return {k: "****" if k in obfuscate_params else v for k, v in params.items()}
    return params


def _execute(
    c: _SnowflakeCursor,
    command: str,
    *,
    kwargs: dict[str, Any] | None = None,
    logger: logging.Logger | None = None,
    command_id: str | None = None,
    obfuscate_params: bool | str | list[str] = False,
    level: int | str = logging.DEBUG,
) -> _SnowflakeCursor | dict[str, Any] | None:
    """Logs the command and executes it with the Snowflake cursor.

    Args:
        c (SnowflakeCursor): The Snowflake cursor or connection to use.
        command (str): The command to execute.
        kwargs (dict[str, Any] | None, optional): The kwargs to pass to the `c.execute`. Defaults to None.
        logger (logging.Logger | None): The logger to use.
        command_id (str, optional): The ID to use in the logs. Generates a random one if not provided. Defaults to None.
        obfuscate_params (bool | str | list[str], optional): Whether to obfuscate the parameters in the logs.
            If `True`, obfuscates all parameters. If a string or list of strings, obfuscates only the specified parameters.
            Defaults to False.
        level (int | str, optional): The logging level to use. Defaults to logging.DEBUG.

    Returns:
        Any: The result of the `c.execute`.

    Example:

        ```python
        import logging
        from snowflake.connector import connect
        from prefecto.ext.snowflake import _execute

        conn = connect(...)
        logger = logging.getLogger(__name__)

        _execute(conn.cursor(), "SELECT * FROM table", logger=logger)
        ```

        Produces logs like

        ```txt
        INFO - [rustic-human] Beginning command.
        DEBUG - [rustic-human] Executing command:
        SELECT * FROM table
        INFO - [rustic-human] Command executed successfully.
        ```

        And tracebacks like

        ```txt
        ERROR - [rustic-human] Command failed.
        Traceback (most recent call last):
        ...
        snowflake.connector.errors.ProgrammingError: ...
        [rustic-human] Command failed
        SELECT * FROM table
        ```

        You can also obfuscate parameters:

        ```python
        _execute(conn.cursor(), "SELECT * FROM table WHERE id = %(id)s", kwargs=dict(params={"id": 123}), obfuscate_params=True)
        _execute(conn.cursor(), "SELECT * FROM table WHERE id = %(id)s AND %(secret)s", kwargs=dict(params={"id": 123, "secret": "shhh"}), obfuscate_params=["secret"])
        ```

        ```txt
        INFO - [delightful-octopus] Beginning command.
        DEBUG - [delightful-octopus] Executing command:
        SELECT * FROM table WHERE id = ****
        INFO - [delightful-octopus] Command executed successfully.
        INFO - [super-athlete] Beginning command.
        DEBUG - [super-athlete] Executing command:
        SELECT * FROM table WHERE id = 123 AND secret = ****
        INFO - [super-athlete] Command executed successfully.
        ```

    """
    kwargs = kwargs or {}
    level_int = logging._nameToLevel.get(level) if isinstance(level, str) else level
    assert isinstance(
        level_int, int
    ), f"Provided logging level {level} didn't convert to int ({level_int}, {type(level_int).__name__})"

    command_id = command_id or "-".join(coolname.generate(2))
    logger = CommandLogAdapter(
        logger or logging.getLogger(__name__), extra={"command_id": command_id}
    )

    # Obfuscate parameters if requested
    formatted_command_string = command % _obfuscate_params(
        kwargs.get("params", {}), obfuscate_params
    )

    logger.info("Beginning command.")
    try:
        logger.log(
            level_int,
            f"Executing command:\n{formatted_command_string}",
        )
        result = c.execute(command, **kwargs)
    except Exception as e:
        if hasattr(e, "add_note") and callable(e.add_note):
            e.add_note(f"Command failed\n{formatted_command_string}")
        logger.error("Command failed.", exc_info=e, stack_info=True)
        raise e
    logger.info("Command executed successfully.")
    return result


def execute(
    c: _SnowflakeCursor,
    command: str,
    params: Sequence[Any] | dict[Any, Any] | None = None,
    _bind_stage: str | None = None,
    timeout: int | None = None,
    _exec_async: bool = False,
    _no_retry: bool = False,
    _do_reset: bool = True,
    _put_callback: SnowflakeProgressPercentage = None,
    _put_azure_callback: SnowflakeProgressPercentage = None,
    _put_callback_output_stream: IO[str] = sys.stdout,
    _get_callback: SnowflakeProgressPercentage = None,
    _get_azure_callback: SnowflakeProgressPercentage = None,
    _get_callback_output_stream: IO[str] = sys.stdout,
    _show_progress_bar: bool = True,
    _statement_params: dict[str, str] | None = None,
    _is_internal: bool = False,
    _describe_only: bool = False,
    _no_results: bool = False,
    _is_put_get: bool | None = None,
    _raise_put_get_error: bool = True,
    _force_put_overwrite: bool = False,
    file_stream: IO[bytes] | None = None,
    *,
    logger: logging.Logger | None = None,
    command_id: str | None = None,
    obfuscate_params: bool | str | list[str] = False,
    level: int | str = logging.DEBUG,
) -> _SnowflakeCursor | dict[str, Any] | None:
    """Executes a command on the Snowflake connection after logging it.

    Args:
        c (SnowflakeCursor): The Snowflake cursor or connection to use.
        command (str): The command to execute.
        params (Sequence[Any] | dict[Any, Any] | None, optional): The parameters to pass to the `c.execute`. Defaults to None.
        _bind_stage (str | None, optional): The stage to bind. Defaults to None.
        timeout (int | None, optional): The timeout to use. Defaults to None.
        _exec_async (bool, optional): Whether to execute asynchronously. Defaults to False.
        _no_retry (bool, optional): Whether to retry on failure. Defaults to False.
        _do_reset (bool, optional): Whether to reset the connection. Defaults to True.
        _put_callback (SnowflakeProgressPercentage, optional): The callback for PUT operations. Defaults to None.
        _put_azure_callback (SnowflakeProgressPercentage, optional): The callback for Azure PUT operations. Defaults to None.
        _put_callback_output_stream (IO[str], optional): The output stream for PUT callbacks. Defaults to sys.stdout.
        _get_callback (SnowflakeProgressPercentage, optional): The callback for GET operations. Defaults to None.
        _get_azure_callback (SnowflakeProgressPercentage, optional): The callback for Azure GET operations. Defaults to None.
        _get_callback_output_stream (IO[str], optional): The output stream for GET callbacks. Defaults to sys.stdout.
        _show_progress_bar (bool, optional): Whether to show a progress bar. Defaults to True.
        _statement_params (dict[str, str] | None, optional): The statement parameters. Defaults to None.
        _is_internal (bool, optional): Whether the command is internal. Defaults to False.
        _describe_only (bool, optional): Whether to describe only. Defaults to False.
        _no_results (bool, optional): Whether to return no results. Defaults to False.
        _is_put_get (bool | None, optional): Whether the command is a PUT or GET operation. Defaults to None.
        _raise_put_get_error (bool, optional): Whether to raise an error on PUT or GET failure. Defaults to True.
        _force_put_overwrite (bool, optional): Whether to force PUT overwrite. Defaults to False.
        file_stream (IO[bytes] | None, optional): The file stream to use. Defaults to None.
        logger (logging.Logger | None): The logger to use.
        command_id (str, optional): The ID to use in the logs. Generates a random one if not provided. Defaults to None.
        obfuscate_params (bool | str | list[str], optional): Whether to obfuscate the parameters in the logs.
            If `True`, obfuscates all parameters. If a string or list of strings, obfuscates only the specified parameters.
            Defaults to False.
        level (int | str, optional): The logging level to use. Defaults to logging.DEBUG.

    Returns:
        SnowflakeCursor | dict[str, Any] | None: The result of the `c.execute`.

    Example:

        ```python
        import logging
        from snowflake.connector import connect
        from prefecto.ext.snowflake import execute

        c = connect(...).cursor()
        execute(c, "SELECT * FROM table")
        ```

        Produces logs like

        ```txt
        INFO - [rustic-human] Beginning command.
        DEBUG - [rustic-human] Executing command:
        SELECT * FROM table
        INFO - [rustic-human] Command executed successfully.
        ```

        And tracebacks like

        ```txt
        ERROR - [rustic-human] Command failed.
        Traceback (most recent call last):
        ...
        snowflake.connector.errors.ProgrammingError: ...
        [rustic-human] Command failed
        SELECT * FROM table
        ```

        You can also obfuscate parameters:

        ```python
        execute(c, "SELECT * FROM table WHERE id = %(id)s", params={"id": 123}, obfuscate_params=True)
        execute(c, "SELECT * FROM table WHERE id = %(id)s AND %(secret)s", params={"id": 123, "secret": "shhh"}, obfuscate_params=["secret"])
        ```

        ```txt
        INFO - [delightful-octopus] Beginning command.
        DEBUG - [delightful-octopus] Executing command:
        SELECT * FROM table WHERE id = ****
        INFO - [delightful-octopus] Command executed successfully.
        INFO - [super-athlete] Beginning command.
        DEBUG - [super-athlete] Executing command:
        SELECT * FROM table WHERE id = 123 AND secret = ****
        INFO - [super-athlete] Command executed successfully.
        ```

    """

    return _execute(
        c,
        command,
        kwargs={
            "params": params,
            "_bind_stage": _bind_stage,
            "timeout": timeout,
            "_exec_async": _exec_async,
            "_no_retry": _no_retry,
            "_do_reset": _do_reset,
            "_put_callback": _put_callback,
            "_put_azure_callback": _put_azure_callback,
            "_put_callback_output_stream": _put_callback_output_stream,
            "_get_callback": _get_callback,
            "_get_azure_callback": _get_azure_callback,
            "_get_callback_output_stream": _get_callback_output_stream,
            "_show_progress_bar": _show_progress_bar,
            "_statement_params": _statement_params,
            "_is_internal": _is_internal,
            "_describe_only": _describe_only,
            "_no_results": _no_results,
            "_is_put_get": _is_put_get,
            "_raise_put_get_error": _raise_put_get_error,
            "_force_put_overwrite": _force_put_overwrite,
            "file_stream": file_stream,
        },
        logger=logger,
        command_id=command_id,
        obfuscate_params=obfuscate_params,
        level=level,
    )


class LogCursor(_SnowflakeCursor):
    """A Snowflake cursor that logs command executions. The `execute` method has additional parameters for altering log behavior.

    Example:

        ```python
        from snowflake.connector import connect
        from prefecto.ext.snowflake import LogCursor

        c = connect(...).cursor(LogCursor)
        c: LogCursor
        c.execute("SELECT * FROM table")
        ```

        Produces logs like

        ```txt
        INFO - [rustic-human] Beginning command.
        DEBUG - [rustic-human] Executing command:
        SELECT * FROM table
        INFO - [rustic-human] Command executed successfully.
        ```
    """

    def execute(
        self,
        command: str,
        params: Sequence[Any] | dict[Any, Any] | None = None,
        _bind_stage: str | None = None,
        timeout: int | None = None,
        _exec_async: bool = False,
        _no_retry: bool = False,
        _do_reset: bool = True,
        _put_callback: SnowflakeProgressPercentage = None,
        _put_azure_callback: SnowflakeProgressPercentage = None,
        _put_callback_output_stream: IO[str] = sys.stdout,
        _get_callback: SnowflakeProgressPercentage = None,
        _get_azure_callback: SnowflakeProgressPercentage = None,
        _get_callback_output_stream: IO[str] = sys.stdout,
        _show_progress_bar: bool = True,
        _statement_params: dict[str, str] | None = None,
        _is_internal: bool = False,
        _describe_only: bool = False,
        _no_results: bool = False,
        _is_put_get: bool | None = None,
        _raise_put_get_error: bool = True,
        _force_put_overwrite: bool = False,
        file_stream: IO[bytes] | None = None,
        *,
        logger: logging.Logger | None = None,
        command_id: str | None = None,
        obfuscate_params: bool | str | list[str] = False,
        level: int | str = logging.DEBUG,
    ) -> _SnowflakeCursor | dict[str, Any] | None:
        """Executes a command on the Snowflake connection after logging it.

        Args:
            command (str): The command to execute.
            params (Sequence[Any] | dict[Any, Any] | None, optional): The parameters to pass to the `c.execute`. Defaults to None.
            _bind_stage (str | None, optional): The stage to bind. Defaults to None.
            timeout (int | None, optional): The timeout to use. Defaults to None.
            _exec_async (bool, optional): Whether to execute asynchronously. Defaults to False.
            _no_retry (bool, optional): Whether to retry on failure. Defaults to False.
            _do_reset (bool, optional): Whether to reset the connection. Defaults to True.
            _put_callback (SnowflakeProgressPercentage, optional): The callback for PUT operations. Defaults to None.
            _put_azure_callback (SnowflakeProgressPercentage, optional): The callback for Azure PUT operations. Defaults to None.
            _put_callback_output_stream (IO[str], optional): The output stream for PUT callbacks. Defaults to sys.stdout.
            _get_callback (SnowflakeProgressPercentage, optional): The callback for GET operations. Defaults to None.
            _get_azure_callback (SnowflakeProgressPercentage, optional): The callback for Azure GET operations. Defaults to None.
            _get_callback_output_stream (IO[str], optional): The output stream for GET callbacks. Defaults to sys.stdout.
            _show_progress_bar (bool, optional): Whether to show a progress bar. Defaults to True.
            _statement_params (dict[str, str] | None, optional): The statement parameters. Defaults to None.
            _is_internal (bool, optional): Whether the command is internal. Defaults to False.
            _describe_only (bool, optional): Whether to describe only. Defaults to False.
            _no_results (bool, optional): Whether to return no results. Defaults to False.
            _is_put_get (bool | None, optional): Whether the command is a PUT or GET operation. Defaults to None.
            _raise_put_get_error (bool, optional): Whether to raise an error on PUT or GET failure. Defaults to True.
            _force_put_overwrite (bool, optional): Whether to force PUT overwrite. Defaults to False.
            file_stream (IO[bytes] | None, optional): The file stream to use. Defaults to None.
            logger (logging.Logger | None): The logger to use.
            command_id (str, optional): The ID to use in the logs. Generates a random one if not provided. Defaults to None.
            obfuscate_params (bool | str | list[str], optional): Whether to obfuscate the parameters in the logs.
                If `True`, obfuscates all parameters. If a string or list of strings, obfuscates only the specified parameters.
                Defaults to False.
            level (int | str, optional): The logging level to use. Defaults to logging.DEBUG.

        Returns:
            SnowflakeCursor | dict[str, Any] | None: The result of the `c.execute`.

        Example:

            ```python
            import logging
            from snowflake.connector import connect
            from prefecto.ext.snowflake import LogCursor

            c = connect(...).cursor(LogCursor)
            c.execute("SELECT * FROM table")
            ```

            Produces logs like

            ```txt
            INFO - [rustic-human] Beginning command.
            DEBUG - [rustic-human] Executing command:
            SELECT * FROM table
            INFO - [rustic-human] Command executed successfully.
            ```

            And tracebacks like

            ```txt
            ERROR - [rustic-human] Command failed.
            Traceback (most recent call last):
            ...
            snowflake.connector.errors.ProgrammingError: ...
            [rustic-human] Command failed
            SELECT * FROM table
            ```

            You can also obfuscate parameters:

            ```python
            c.execute("SELECT * FROM table WHERE id = %(id)s", params={"id": 123}, obfuscate_params=True)
            c.execute("SELECT * FROM table WHERE id = %(id)s AND %(secret)s", params={"id": 123, "secret": "shhh"}, obfuscate_params=["secret"])
            ```

            ```txt
            INFO - [delightful-octopus] Beginning command.
            DEBUG - [delightful-octopus] Executing command:
            SELECT * FROM table WHERE id = ****
            INFO - [delightful-octopus] Command executed successfully.
            INFO - [super-athlete] Beginning command.
            DEBUG - [super-athlete] Executing command:
            SELECT * FROM table WHERE id = 123 AND secret = ****
            INFO - [super-athlete] Command executed successfully.
            ```

        """

        return _execute(
            super(),
            command,
            kwargs={
                "params": params,
                "_bind_stage": _bind_stage,
                "timeout": timeout,
                "_exec_async": _exec_async,
                "_no_retry": _no_retry,
                "_do_reset": _do_reset,
                "_put_callback": _put_callback,
                "_put_azure_callback": _put_azure_callback,
                "_put_callback_output_stream": _put_callback_output_stream,
                "_get_callback": _get_callback,
                "_get_azure_callback": _get_azure_callback,
                "_get_callback_output_stream": _get_callback_output_stream,
                "_show_progress_bar": _show_progress_bar,
                "_statement_params": _statement_params,
                "_is_internal": _is_internal,
                "_describe_only": _describe_only,
                "_no_results": _no_results,
                "_is_put_get": _is_put_get,
                "_raise_put_get_error": _raise_put_get_error,
                "_force_put_overwrite": _force_put_overwrite,
                "file_stream": file_stream,
            },
            logger=logger,
            command_id=command_id,
            obfuscate_params=obfuscate_params,
            level=level,
        )



"""
Tests for the logging module.
"""

from __future__ import annotations

import logging

from prefecto.logging import get_prefect_or_default_logger


def test_get_prefect_or_default_logger():
    """Tests `get_prefect_or_default_logger`."""
    assert get_prefect_or_default_logger().__class__ == logging.RootLogger
    assert logging.getLogger("not root").__class__ == logging.Logger
    assert (
        get_prefect_or_default_logger(logging.Logger("not root")).__class__
        == logging.Logger
    )



"""
Prefect logging utilities.
"""

from __future__ import annotations

import logging

from prefect.logging import get_run_logger


def get_prefect_or_default_logger(
    __default: logging.Logger | str | None = None,
) -> logging.Logger:
    """Gets the Prefect logger if the global context is set. Returns the `__default` or
    root logger if not.
    """
    # Type check the default logger
    if not isinstance(__default, (logging.Logger, str, type(None))):
        raise TypeError(
            f"Expected `__default` to be a `logging.Logger`, `str`, or `None`, "
            f"got `{type(__default).__name__}`."
        )
    try:
        return get_run_logger()
    except RuntimeError:
        if isinstance(__default, str):
            return logging.getLogger(__default)
        return __default or logging.getLogger()
